from astutil.nodes import TreeNode, equals

class Type(TreeNode):        
    def isa(self, other):
##        return self == other
        return equals(self, other)

    def resolve(self, scope):
        return self

    def __str__(self):
        return self.name


class UnresolvedType(Type):
    def resolve(self, scope):
        return scope.resolve(self.name, self.sloc)


class PrimitiveType(Type):
    def __init__(self, **kwargs):
        tag = kwargs.get('tag')
        if tag is None:
            tag = kwargs.get('name') + '-type'
            kwargs['tag'] = tag
        super(PrimitiveType, self).__init__(**kwargs)


class VoidType(PrimitiveType):
    def __init__(self):
        super(VoidType, self).__init__(name='void')

class BoolType(PrimitiveType):
    def __init__(self, **kwargs):
        super(BoolType, self).__init__(name='bool', bitwidth=1, signed=False)

class BitvectorType(PrimitiveType):
    def isa(self, other):
        if isinstance(other, BitvectorType):
            return other.signed == self.signed and other.bitwidth == self.bitwidth
        return False

    def __str__(self):
        return 'bitvector <{}, {}>'.format(self.bitwidth, self.signed)

class IntegralType(BitvectorType):
    pass

class BitType(IntegralType):
    def __init__(self, **kwargs):
        super(BitType, self).__init__(name='bit', bitwidth=1, signed=False)

class CharType(IntegralType):
    def __init__(self, **kwargs):
        super(CharType, self).__init__(name='char', bitwidth=8, signed=True)

class ShortType(IntegralType):
    def __init__(self, **kwargs):
        super(ShortType, self).__init__(name='short', bitwidth=16, signed=True)

class IntType(IntegralType):
    def __init__(self, **kwargs):
        super(IntType, self).__init__(name='int', bitwidth=32, signed=True)

class LongType(IntegralType):
    def __init__(self, **kwargs):
        super(LongType, self).__init__(name='long', bitwidth=64, signed=True)


class UCharType(IntegralType):
    def __init__(self, **kwargs):
        super(UCharType, self).__init__(name='uchar', bitwidth=8, signed=False)

class UShortType(IntegralType):
    def __init__(self, **kwargs):
        super(UShortType, self).__init__(name='ushort', bitwidth=16, signed=False)

class UIntType(IntegralType):
    def __init__(self, **kwargs):
        super(UIntType, self).__init__(name='uint', bitwidth=32, signed=False)

class ULongType(IntegralType):
    def __init__(self, **kwargs):
        super(ULongType, self).__init__(name='ulong', bitwidth=64, signed=False)


class PointerType(Type):
    def __init__(self, **kwargs):
        if not 'tag' in kwargs:
            kwargs['tag'] = 'pointer-type'
        super(PointerType, self).__init__(**kwargs)

    def isa(self, other):
        if not isinstance(other, PointerType):
            return False
        return self.inner_type.isa(other.inner_type)

    def __str__(self):
        return str(self.inner_type) + "*"

class ArrayType(PointerType):
    index_type = ULongType()
    def __init__(self, **kwargs):
        if not 'tag' in kwargs:
            kwargs['tag'] = 'array-type'
        self.length = None # if not given assume default
        super(ArrayType, self).__init__(**kwargs)

    def __str__(self):
        return str(self.inner_type) + '[' + str(self.length) + ']'


# instead of using:
# isa: TypeInstance, TypeClass -> Bool
# use this function, this will simplify the overall logic
def derived(given, expected):
    assert isinstance(given, Type)
    return isinstance(given, expected)

# isa: TypeInstance, TypeInstance -> Bool
# isa: TypeInstance, TypeClass -> Bool
# TypeClass can either be a normal python type (e.g. class) or a user
# defined hierarchy element. TypeInstance is a specific instance of a
# given class, it's TypeClass (whatever it is) can be extracted.
def isa(given, expected):
    assert isinstance(given, Type)
    try:
        # first check type hierarchy
        return derived(given, expected)
    except TypeError:
        assert isinstance(expected, Type)
    # if no type class but a specific type instance is given
    # (e.g. two specific arrays) then use the type specific comparators
    try:
        it = iter(expected)
    except TypeError:
        return given.isa(expected)
    return any((e.isa(expected) for e in it))
