from __future__ import print_function
import sys
import os
import argparse
from astutil.walk import walk_up

import ivl.parser.ivl as ivl
from ivl.parser.output.IVLParser import ParserException
import ivl.analysis.astpasses as astpasses
import ivl.analysis.env as env
import ivl.analysis.report as report

def parse_and_check(inputfile):
    node = None
    try:
        node = ivl.parse_file(inputfile)
        node = astpasses.MapTreeNodes().run(node)
        node = astpasses.ResolveTypes().run(node)
        node = astpasses.CollectLabels().run(node)
        node = astpasses.ResolveDecls().run(node)
        node = astpasses.TypeCheck().run(node)
        return node
    except env.DoubleDeclarationError as e:
        report.add_error(str(e), e.sloc)
        report.add_note("previous declaration was here:", e.prev_sloc)
        report.display_all()
    except astpasses.Error as e:
        report.display_error(str(e), e.sloc)
    except ParserException as e:
        print("failed to parse program")

def simulate(node):
    import ivl.vm.example.interpreter as vm
    try:
        vm.simulate(node)
    except vm.Error as e:
        report.display_error(str(e), e.sloc)

def show_header(s):
    h = '=[ {} ]'.format(s)
    print(h + '='*(80 - len(h)))

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('--run', action='store_true')
    ap.add_argument('inputfile')
    args = ap.parse_args()

    if not os.path.isfile(args.inputfile):
        print("error: file '{}' does not exist".format(args.inputfile))
        sys.exit(-1)

    print("using input file: '{}'".format(args.inputfile))
    prog = parse_and_check(args.inputfile)
    if prog is not None:
        print("program successfully parsed")
        if args.run:
            show_header('simulation')
            simulate(prog)
    else:
        sys.exit(-1)

if __name__ == '__main__':
    main()