grammar IVL;

options {
  language = Python;
}

@header {
from astutil.nodes import *

class ParserException(Exception):
    def __init__(self, message, sloc=None, **kwargs):
        Exception.__init__(self, message)
    	self.sloc = sloc
    	self.__dict__.update(kwargs)

_inputFile = None
_inputText = None

def Token(**kwargs):
    return type('Token', (object,), kwargs)

def sloc(tok):
    if tok is None:
        return None
    return Token(line=tok.getLine(), column=tok.getCharPositionInLine(), text=tok.text, source_file=_inputFile, source_text=_inputText)
    
def text(tok):
    if tok is None:
    	return None
    return tok.text

def VoidType():
    return Type('unresolved-type', None, name='void')
    
def Program():
    return Misc('program', None, funcs=[], threads=[], updates=[], events=[], vars=[])
}

@lexer::members {
# antlr lexer does not count errors for some reason (in contrary to BaseRecognizer) so do it here again
def reportError(self, e):
    self._state.syntaxErrors += 1
    self.displayRecognitionError(self.tokenNames, e)
}

@init {
    self.program = Program()
    self.program.main = None
}


prog [inputFile, inputText] returns [val]
@init {
  global _inputFile
  global _inputText
  _inputFile = inputFile
  _inputText = inputText
}
	:	( global_var
		| event 
		| thread 
		| update 
		| func 
		| nl
		| main
		)* {
			$val = self.program._rebuilt_children()
		} EOF
	;
	
nl	:	NEWLINE
	;
	
	
global_var returns [val]
	:	var	{
			$val = $var.val
			$val.loc = 'global'
			self.program.vars.append($var.val)
		}
	;
	
local_var returns [val]
	:	var	{
			$val = $var.val
			$val.loc = 'local'
		}
	;
	
param_var returns [val]
	:	var	{
			$val = $var.val
			$val.loc = 'param'
		}
	;
	
var returns [val]
@init {
  is_array = False
}
 	:	type ID ('[' (index=expr)? ']' {is_array=True})? ('=' rhs=expr)?	{
 			vartype = $type.val
 			if is_array:
	 			vartype = Type('array-type', sloc($ID), inner_type=vartype, length=$index.val)
	 		$val = Decl('var-decl', sloc($ID), name=$ID.text, type=vartype, initializer=$rhs.val, is_array=is_array)
 		}
	;	
	
event	:	'event' ID	{ 
			self.program.events.append( Decl('event-decl', sloc($ID), name=$ID.text) )
		}
	;	

thread	:	'thread' ID body	{
			fn = Decl('thread-decl', sloc($ID), name=$ID.text, body=$body.val, params=[], return_type=VoidType())
			self.program.threads.append( fn )
		}
	;
	
update	:	'update' ID body	{
			fn = Decl('update-decl', sloc($ID), name=$ID.text, body=$body.val, params=[], return_type=VoidType())
			self.program.updates.append( fn )
		}
	;
	
func	:	type ID '(' formal_args ')' body	{
			fn = Decl('func-decl', sloc($ID), name=$ID.text, body=$body.val, params=$formal_args.val, return_type=$type.val)
			self.program.funcs.append( fn )
		}
	;
	
type returns [val]
	:	primitive_type	{
			$val = $primitive_type.val
		}
		(('*') => tok='*'	{
			$val = Type('pointer-type', sloc($tok), inner_type=$val)
		})*
	;
	
primitive_type returns [val]
	:	tok='signed' '<' INT '>'	{
			$val = Type('bitvector-type', sloc($tok), bitwidth=int($INT.text), signed=True)
		}
	|	tok='unsigned' '<' INT '>'	{
			$val = Type('bitvector-type', sloc($tok), bitwidth=int($INT.text), signed=False)
		}
	|	ID	{
			$val = Type('unresolved-type', sloc($ID), name=$ID.text)
		}
	;
	
formal_args returns [val]
@init {
  $val = []
}
	:	(v1=param_var {$val.append($v1.val);} (',' v2=param_var {$val.append($v2.val);})*)?
	;
	
main	:	tok='main' body	{
			fn = Decl('func-decl', sloc($tok), name='main', body=$body.val, params=[], return_type=VoidType())
			if self.program.main is not None:
				raise ParserException("main function can only be declared once", sloc($tok))
			self.program.main = fn
		}
	;

	
body returns [val]
	:	tok='begin' nl+ stmt_list	{
			$val = Stmt('body', sloc($tok), stmts=$stmt_list.val)
		} 'end'
	;
	

stmt returns [val]
	:	tok='print' expr		{ $val = Stmt('print', sloc($tok), value=$expr.val) }
	|	tok='puts' STRING		{ $val = Stmt('show', sloc($tok), message=$STRING.text) }
	|	tok='assert' expr STRING?	{ $val = Stmt('assert', sloc($tok), cond=$expr.val, message=text($STRING)) }
	|	tok='assume' expr		{ $val = Stmt('assume', sloc($tok), cond=$expr.val) }
	|	tok='suspend' ID		{ $val = Stmt('suspend-thread', sloc($tok), thread_name=$ID.text) }
	|	tok='resume' ID			{ $val = Stmt('resume-thread', sloc($tok), thread_name=$ID.text) }
	|	tok='return' expr?		{ $val = Stmt('return', sloc($tok), value=$expr.val) }
	|	tok='request_update' th=ID	{ $val = Stmt('request-update', sloc($tok), update_name=$th.text) }
	|	tok='dont_initialize' th=ID events=id_list	{ 
			$val = Stmt('dont-initialize', sloc($tok), thread_name=$th.text, event_names=$id_list.val)
		}
	|	(tok='wait_time' | tok='delay') expr	{ 
			$val = Stmt('wait-time', sloc($tok), delay=$expr.val)
		}
	|	(tok='wait_event' | tok='wait') ID 	{
			$val = Stmt('wait-event', sloc($tok), event_name=$ID.getText())
		}
	|	tok='notify' ID expr?	{ 
			$val = Stmt('notify-event', sloc($tok), event_name=$ID.getText(), delay=$expr.val)
		}
	|	tok='goto' ID		{
			$val = Stmt('goto', sloc($tok), label_name=$ID.getText())
		}
	|	ID tok=':'		{ 
			$val = Stmt('label', sloc($tok), label_name=$ID.text)
		}
	|	(local_var) => local_var	{
			$val = $local_var.val
		}
	|	expr {
			$val = Stmt('eval', $expr.val.sloc, expr=$expr.val)
		}
	|	tok='if' expr 'goto' ID	{
			$val = Stmt('if', sloc($tok), cond=$expr.val, goto=Stmt('goto', sloc($ID), label_name=$ID.text))
		}
	|	tok='start' expr?	{
			$val = Stmt('start-sim', sloc($tok), steps=$expr.val)
		}
	|	tok='stop'	{
			$val = Stmt('stop-sim', sloc($tok))
		}
	;	
	
stmt_list returns [val]
@init {
  $val = [];
}
	:	(stmt {$val.append($stmt.val);} nl+)*
	;	
	


expr returns [val]
	:	assignment	{ $val = $assignment.val; }
	;
		
assignment returns [val]
@init {
def compound_assign(tok, op, lhs, rhs):
    token = sloc(tok)
    if op != None:
        rhs = Expr('binary', token, op=op, lhs=lhs, rhs=rhs)
    return Expr('assign', token, lhs=lhs, rhs=rhs)
}
	:	lhs=disjunction {$val = $lhs.val;} (
			( tok='='   {op = None}
			| tok='+='  {op = '+'}
			| tok='-='  {op = '-'}
			| tok='|='  {op = '|'}
			| tok='&='  {op = '&'}
			| tok='^='  {op = '^'}
			| tok='>>=' {op = '>>'}
			| tok='<<=' {op = '<<'}
			) rhs=disjunction {$val = compound_assign($tok, op, $val, $rhs.val);})*
	;
		
disjunction returns [val]
	:	lhs=conjunction {$val = $lhs.val;} (tok='||' rhs=conjunction {
			$val = Expr('binary', sloc($tok), op='||', lhs=$val, rhs=$rhs.val);})*
	;
	
conjunction returns [val]
	:	lhs=bitwise_or {$val = $lhs.val;} (tok='&&' rhs=bitwise_or {
			$val = Expr('binary', sloc($tok), op='&&', lhs=$val, rhs=$rhs.val);})*
	;
	
bitwise_or returns [val]
	:	lhs=bitwise_xor {$val = $lhs.val;} (tok='|' rhs=bitwise_xor {
			$val = Expr('binary', sloc($tok), op='|', lhs=$val, rhs=$rhs.val);})*
	;
	
bitwise_xor returns [val]
	:	lhs=bitwise_and {$val = $lhs.val;} (tok='^' rhs=bitwise_and {
			$val = Expr('binary', sloc($tok), op='^', lhs=$val, rhs=$rhs.val);})*
	;
	
bitwise_and returns [val]
	:	lhs=relation {$val = $lhs.val;} (tok='&' rhs=relation {
			$val = Expr('binary', sloc($tok), op='&', lhs=$val, rhs=$rhs.val);})*
	;
	
relation returns [val]
	:	lhs=ordrelation {$val = $lhs.val;} (
			( tok='=='
			| tok='!='
			) rhs=ordrelation {$val = Expr('binary', sloc($tok), op=tok.text, lhs=$val, rhs=$rhs.val);})*
	;	
	
ordrelation returns [val]
	:	lhs=bit_shift {$val = $lhs.val;} (
			( tok='<'
			| tok='<='
			| tok='>'
			| tok='>='
			) rhs=bit_shift {$val = Expr('binary', sloc($tok), op=tok.text, lhs=$val, rhs=$rhs.val);})*
	;

bit_shift returns [val]
	:	lhs=term {$val = $lhs.val;} (
			( tok='<<'
			| tok='>>'
			) rhs=term {$val = Expr('binary', sloc($tok), op=text, lhs=$val, rhs=$rhs.val);})*
	;
	
term returns [val]
	:	lhs=factor {$val = $lhs.val;} (
			( tok='+'
			| tok='-'
			) rhs=factor {$val = Expr('binary', sloc($tok), op=tok.text, lhs=$val, rhs=$rhs.val);})*
	;		
	
factor returns [val]
	:	lhs=unary {$val = $lhs.val;} (
			( tok='*'
			| tok='/'
			| tok='\%'
			) rhs=unary {$val = Expr('binary', sloc($tok), op=tok.text, lhs=$val, rhs=$rhs.val);})*
	;
	
unary returns [val]
	:	tok='-' e=unary		{ $val = Expr('unary', sloc($tok), op='-', operand=$e.val) }
	|	tok='!' e=unary		{ $val = Expr('unary', sloc($tok), op='!', operand=$e.val) }
	|	tok='~' e=unary		{ $val = Expr('unary', sloc($tok), op='~', operand=$e.val) }
	|	tok='*' e=unary		{ $val = Expr('deref', sloc($tok), expr=$e.val) }
	|	tok='&' e=unary		{ $val = Expr('addrof', sloc($tok), expr=$e.val) }
	|	tok='length' e=unary	{ $val = Expr('length', sloc($tok), array=$e.val) }
	|	new_or_delete		{ $val = $new_or_delete.val }
	|	('(' type ')' unary) => tok='(' type ')' e=unary { $val = Expr('cast', sloc($tok), expr=$e.val, target_type=$type.val) }
	|	access			{ $val = $access.val }
	;
	
access returns [val]
	:	array=atom {
			$val = $array.val
		}
		(tok='[' index=expr ']'	{
			$val = Expr('access', sloc($tok), array=$val, index=$index.val)
		})?
	;

atom returns [val]
	:	number		{ $val = Expr('number', sloc($number.tok), value=$number.val) }
	|	ID		{ $val = Expr('id', sloc($ID), name=$ID.text, value=$ID.text) }
	|	CHAR		{ $val = Expr('char', sloc($CHAR), value=ord($CHAR.text)) }
	|	STRING		{ $val = Expr('string', sloc($STRING), value=$STRING.text) }
	|	tok='false'	{ $val = Expr('bool', sloc($tok), value=False) }
	|	tok='true'	{ $val = Expr('bool', sloc($tok), value=True) }
	|	tok='?' ID? '<' type '>'	{ $val = Expr('symbolic', sloc($tok), name=text($ID), type=$type.val) }
	|	call		{ $val = $call.val }
	|	'(' expr ')' 	{ $val = $expr.val }
	;
	
new_or_delete returns [val]
@init {
  is_array = False
}
	:	tok='new' type ('[' expr ']' {is_array=True})?	{
			type = $type.val
			if is_array:
				type = Type('array-type', sloc($tok), inner_type=type, length=$expr.val)
			else:
				type = Type('pointer-type', sloc($tok), inner_type=type)
			$val = Expr('new', sloc($tok), type=type, length=$expr.val, is_array=is_array)
		}
	|	tok='delete' ('['']' {is_array=True})? access	{ 
			$val = Expr('delete', sloc($tok), expr=$access.val, is_array=is_array) 
		}
	;
	
call returns [val]
	:	ID (args { $val = Expr('call', sloc($ID), name=$ID.text, args=$args.val); })
	;
	
args returns [val]
@init {
  $val = []
}
	:	 ('(' (e1=expr {$val.append($e1.val);})? (',' e2=expr {$val.append($e2.val);})* ')')
	;
	
expr_list returns [val]
@init {
  $val = []
}
	:	e1=expr { $val.append($e1.val); } (',' e2=expr { $val.append($e2.val); })*
	;
	
id_list returns [val]
@init {
  $val = []
}
	:	e1=ID { $val.append($e1.text) } (e2=ID { $val.append($e2.text) })*
	;

number returns [val, tok]
	:	INT	{ 
			$val = int($INT.text, 10)
			$tok = $INT
		}
	;


ID  :	('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*
    ;

INT :	'0'..'9'+
    ;

COMMENT
    :   '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    |   '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

WS  :   ( ' '
        | '\t'
        ) {$channel=HIDDEN;}
    ;
    
NEWLINE	:	'\n' | '\r'
	;

STRING
    :  '"' ( ESC_SEQ | ~('\\'|'"') )* '"'	{ self.text=self.text[1:-1]; }
    ;
    
CHAR
    :  '\'' ( ESC_SEQ | ~('\\'|'\'') ) '\''	{ self.text=self.text[1:-1]; }
    ;

fragment
HEX_DIGIT : ('0'..'9'|'a'..'f'|'A'..'F') ;

fragment
BIN_DIGIT : ('0' | '1') ;

fragment
ESC_SEQ
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   UNICODE_ESC
    |   OCTAL_ESC
    ;

fragment
OCTAL_ESC
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;

fragment
UNICODE_ESC
    :   '\\' 'u' HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT
    ;
