from peak.rules import when, abstract
from bitstring import BitArray
from collections import defaultdict
from random import randrange
from ivl.analysis.coercion import is_array, is_pointer, is_bitvector, is_bool


class Error(Exception):
    pass

class SimulationError(Error):
    def __init__(self, what, sloc):
        self.what = what
        self.sloc = sloc
        super(SimulationError, self).__init__("simulation failed with: {}".format(what))

def raise_error(message, sloc):
    raise SimulationError(message, sloc)
    

# Container values for each type

HEAP, STACK, READONLY, GLOBAL = range(4)

class Value(object):
    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.__dict__)
    def _unsupported_operation(self, name, sloc):
        raise_error("operation '{}' not supported for '{}'".format(name, repr(self)), sloc)

    def read(self, sloc): self._unsupported_operation('read', sloc)
    def write(self, value, sloc): self._unsupported_operation('write', sloc)
    def deref(self, sloc): self._unsupported_operation('deref', sloc)
    def index(self, idx, sloc): self._unsupported_operation('index', sloc)
    def length(self, vm, idx_type, sloc): self._unsupported_operation('length', sloc)
    def addrof(self, sloc): self._unsupported_operation('addrof', sloc)
    def delete(self, sloc): self._unsupported_operation('delete', sloc)
    def delete_array(self, sloc): self._unsupported_operation('delete_array', sloc)

class Undefined(Value): pass

class Bitvector(Value):
    def __init__(self, value, bv_type):
        self.signed = bv_type.signed
        self.bitwidth = bv_type.bitwidth
        if self.signed:
            self.content = BitArray(int=value, length=self.bitwidth)
        else:
            self.content = BitArray(uint=value, length=self.bitwidth)

    def native_value(self):
        return self.content.int if self.signed else self.content.uint

class Bool(Value):
    def __init__(self, value):
        self.content = value

    def native_value(self):
        return self.content

class Array(Value):
    def __init__(self, elements):
        self.elements = elements

    def index(self, idx, sloc):
        if idx < 0 or idx >= len(self.elements):
            raise_error("array out of bounds error", sloc)
        return self.elements[idx]

    def length(self, vm, idx_type, sloc):
        return construct(vm, sloc, idx_type, READONLY, len(self.elements))

    def deref(self, sloc):
        return self.index(0, sloc)

    def delete_array(self, sloc):
        # NOTE: see Pointer.delete
        for ref in self.elements:
            ref.destroy(sloc)

class Pointer(Value):
    def __init__(self, sloc, target):
        self.sloc = sloc
        self.target = target

    def _not_undefined(fn):
        def check(self, sloc):
            if type(self.target) == Undefined:
                raise_error('attempt to access undefined poiner', sloc)
            return fn(self, sloc)
        return check

    @_not_undefined
    def deref(self, sloc):
        return self.target

    @_not_undefined
    def delete(self, sloc):
        # NOTE: check the location of this object and delete it if the
        # operation is valid (e.g. location is heap and not deleted already).
        # Recursively mark all reachable objects as destroyed (actually
        # destroy them bottom up, e.g. if destructors of structs should be
        # called)
        self.target.destroy(sloc)

class Storage(Value):
    def __init__(self, sloc, value, location):
        self.sloc = sloc
        self.value = value
        self.destroyed = False
        self.location = location

    def _not_destroyed(fn):
        def check(self, *args):
            if self.destroyed:
                raise_error('attempt to access deallocated location', args[-1])
            return fn(self, *args)
        return check

    @_not_destroyed
    def read(self, sloc):
        if type(self.value) == Undefined:
            raise_error("attempt to read an uninitialized location", sloc)
        return self.value

    @_not_destroyed
    def write(self, value, sloc):
        self.value = value

    @_not_destroyed
    def deref(self, sloc):
        return self.value.deref(sloc)

    def addrof(self, sloc):
        return Pointer(sloc, self)

    def destroy(self, sloc):
        if self.location != HEAP:
            raise_error("can only deallocate objects on the heap", sloc)
        if self.destroyed:
            raise_error("double deallocation error", sloc)
        self.destroyed = True


@abstract
def construct(vm, sloc, type, location, initval=None): pass

@when(construct, "is_bitvector(type)")
def construct(vm, sloc, type, location, initval=None):
    if initval is None:
        if location == GLOBAL:
            initval = 0
        else:
            return Undefined()
    return Bitvector(initval, type)

@when(construct, "is_bool(type)")
def construct(vm, sloc, type, location, initval=None):
    if initval is None:
        if location == GLOBAL:
            initval = False
        else:
            return Undefined()
    return Bool(initval)

@when(construct, "is_array(type)")
def construct(vm, sloc, type, location, initval=None):
    def field(e):
        return Storage(sloc, construct(vm, sloc, type.inner_type, location, e), location)
    if initval is not None:
        return Array([field(e) for e in initval])
    assert type.length is not None
    size = vm._eval(type.length).native_value()
    return Array([field(None) for _ in xrange(size)])

@when(construct, "is_pointer(type) and not is_array(type)")
def construct(vm, sloc, type, location, initval=None):
    assert(initval is None) # assume for now
    return Pointer(sloc, Undefined())


TERMINATED, BLOCKED = range(2)

"""
 The given interpreter should provide:

  def execute_non_blocking(fn, args)
  def continue_execution(thread)

Where execute_non_blocking will run the given function (in this case
an update function will be used) with the provided arguments. If any
blocking statement is evaluated during this functions execution an error
should be raised.
The resume function is used to start or continue executing a thread so it
takes no arguments. It will return the new status of the thread after execution,
this can be either of: TERMINATED, BLOCKED.
"""
class Kernel(object):
    def __init__(self, threads, vm):
        self.vm         = vm           # interpreter that implements the necessary interface functions
        self.threads    = set(threads) # all threads
        self.running    = None         # currently running thread
        self.suspended  = set()        # set of suspended threads
        self.waiting    = defaultdict(set)       # event -> set of threads
        self.pending    = defaultdict(set)       # time -> set of events
        self.delayed    = defaultdict(set)       # time -> set of threads
        self.terminated = set()        # set of terminated threads
        self.to_update  = set()        # functions for which request_update has been issued
        self.sim_time   = 0

    @property
    def runnable(self):
        return set.difference(self.threads,
                              set.union(
                                  set.union(*(self.waiting.values()+[set()])),
                                  set.union(*(self.delayed.values()+[set()])),
                                  self.suspended,
                                  self.terminated))

    def _advance_time(self, n):
        self.sim_time += n
        events = self.pending.pop(n, [])
        for e in events:
            self.waiting.pop(e, None)
        self.delayed.pop(n, None)

    def _eval_phase(self):
        while len(self.runnable) > 0:
            self.running = self.runnable.pop()
            stat = self.vm.execute_thread(self.running)
            if stat == TERMINATED:
                self.terminated.add(self.running)
        self.running = None

    def _update_phase(self):            
        for fn in self.to_update:
            self.vm.execute_non_blocking(fn, [])
        self.to_update = set()

    def _delta_notify_phase(self):
        self._advance_time(0)

    def _timed_notify_phase(self):
        times = self.pending.keys() + self.delayed.keys()
        if len(times) > 0:
            n = min(times)
            self._advance_time(n)

    # provided interface functions (all of them take already constant
    # values - no need to evaluate them since that's task of the interpreter)

    def simulation_cycle(self, max_time):
        while len(self.runnable) > 0 and self.sim_time <= max_time:
            self._eval_phase()
            self._update_phase()
            self._delta_notify_phase()
            if len(self.runnable) == 0:
                self._timed_notify_phase()

    def immediate_notify(self, event):
        self.waiting.pop(event, None)

    def timed_notify(self, event, delay):
        assert(delay >= 0)
        self.pending[delay].add(event)

    def wait_time(self, delay):
        assert(delay >= 0)
        self.delayed[delay].add(self.running)

    def wait_event(self, event):
        self.waiting[event].add(self.running)

    def request_update(self, func):
        self.to_update.add(func)

    def suspend(self, thread):
        self.suspended.add(thread)

    def resume(self, thread):
        self.suspended.discard(thread)


class ThreadBlocked(Exception):
    def __init__(self, thread, sloc):
        self.thread = thread
        self.sloc   = sloc

class CallFrame(object):
    def __init__(self):
        self.ip = 0         # instruction pointer
        self.env = {}       # local variables
        self.retval = None  # propagate return value on this custom stack since the internal python stack will be destroyed when a function is blocked

class Interpreter(object):
    def __init__(self, program):
        self.program = program
        self.kernel  = Kernel(program.threads, self)
        self.global_env = dict()  # var -> value
        self.callstacks = dict()  # thread/main -> callstack

    def _get_ip(self):
        return self.callstacks[self.running][-1].ip

    def _set_ip(self, ip):
        self.callstacks[self.running][-1].ip = ip

    ip = property(_get_ip, _set_ip)

    def _get_local_env(self):
        return self.callstacks[self.running][-1].env

    local_env = property(_get_local_env)

    def _enter_block(self):
        if self.running not in self.callstacks:
            self.callstacks[self.running] = []
        self.callstacks[self.running].append(CallFrame())

    def _leave_block(self):
        self.callstacks[self.running].pop()
        if len(self.callstacks[self.running]) == 0:
            del self.callstacks[self.running]

    @property
    def running(self):
        if self.kernel.running is None:
            return self.program.main
        else:
            return self.kernel.running


    def run(self):
        for v in self.program.vars:
            self._eval(v)
        self.execute_non_blocking(self.program.main, [])

    def _eval_start(self):
        self.kernel.simulation_cycle()

    def execute_thread(self, thread):
        try:
            if thread in self.callstacks:
                self._eval_body(thread.body)
            else:
                self._eval_function(thread, [])
            return TERMINATED
        except ThreadBlocked:
            return BLOCKED

    def execute_non_blocking(self, fn, args):
        try:
            return self._eval_function(fn, args)
        except ThreadBlocked as e:
            raise_error("not allowed to block here", e.sloc)

    def _eval(self, node):
        return getattr(self, '_eval_' + node.tag.replace('-', '_'))(node)


    def _eval_function(self, fn, args):
        self._enter_block()
        # assign actual parameters
        for idx, var in enumerate(fn.params):
            self.local_env[var.name] = Storage(var.sloc, args[idx], STACK)
        # execute function body
        val = self._eval(fn.body)
        self._leave_block()
        return val

    def _eval_body(self, node):
        val = None
        while self.ip < len(node.stmts):
            stmt = node.stmts[self.ip]
            self.ip += 1
            val = self._eval(stmt)
        return val

    def _eval_call(self, node):
        return self.execute_non_blocking(
            node.callee,
            [self._eval(expr) for expr in node.args])

    def _eval_return(self, node):
        self.ip = len(node.func.body.stmts)
        if node.value is not None:
            return self._eval(node.value)

    def _eval_goto(self, node):
        self.ip = node.label.jump_index

    def _eval_label(self, node):
        pass  # nothing to do here

    def _eval_if(self, node):
        x = self._eval(node.cond)
        if x.native_value():
            self._eval(node.goto)

    def _eval_eval(self, node):
        return self._eval(node.expr)


    def _eval_number(self, node):
        return construct(self, node.sloc, node.type, READONLY, node.value)

    def _eval_char(self, node):
        return construct(self, node.sloc, node.type, READONLY, node.value)

    def _eval_string(self, node):
        return construct(self, node.sloc, node.type, READONLY, [ord(c) for c in node.value])

    def _eval_bool(self, node):
        return construct(self, node.sloc, node.type, READONLY, node.value)

    def _eval_symbolic(self, node):
        # NOTE: handle symbolic values instead using random value
        return construct(self, node.sloc, node.type, READONLY, randrange(0, 2**(node.type.bitwidth-1)))

    def _eval_id(self, node):
        if node.decl.loc == 'global':
            return self.global_env[node.name]
        return self.local_env[node.name]

    def _eval_var_decl(self, node):
        loc = GLOBAL if node.loc == 'global' else STACK
        if node.initializer is None:
            val = construct(self, node.sloc, node.type, loc)
        else:
            val = self._eval(node.initializer)
        ref = Storage(node.sloc, val, loc)
        if node.loc == 'global':
            self.global_env[node.name] = ref
        else:
            self.local_env[node.name] = ref
        return ref

    def _eval_slice(self, node):
        val = self._eval(node.expr).native_value()
        return construct(self, node.sloc, node.type, READONLY, val)

    def _eval_extend(self, node):
        return self._eval_slice(node)

    def _eval_bitvector_to_bool(self, node):
        val = self._eval(node.expr).native_value() == 0
        return construct(self, node.sloc, node.type, READONLY, val)

    def _eval_bool_to_bitvector(self, node):
        val = 1 if self._eval(node.expr).native_value() else 0
        return construct(self, node.sloc, node.type, READONLY, val)

    def _eval_bool_to_integral(self, node):
        # each integral is a bitvector so it can be handled by the same function
        return self._eval_bool_to_bitvector(node)

    def _eval_cast(self, node):
        raise_error('not implemented', node.sloc)

    def _eval_binary(self, node):
        # NOTE: handle pointer operations if necessary
        assert(not is_pointer(node.lhs) and not is_pointer(node.rhs))
        lhs = self._eval(node.lhs).native_value()
        rhs = self._eval(node.rhs).native_value()
        result = eval('{} {} {}'.format(lhs, node.op, rhs))
        return construct(self, node.sloc, node.type, READONLY, result)

    def _eval_unary(self, node):
        arg = self._eval(node.operand).native_value()
        try:
            op = {
                '!': 'not'
            }[node.op]
        except KeyError:
            op = node.op
        result = eval('{} {}'.format(op, arg))
        return construct(self, node.sloc, node.type, READONLY, result)

    def _eval_assign(self, node):
        ref = self._eval(node.lhs)
        val = self._eval(node.rhs)
        ref.write(val, node.sloc)
        return ref

    def _eval_valueof(self, node):
        ref = self._eval(node.expr)
        return ref.read(node.sloc)

    def _eval_deref(self, node):
        ref = self._eval(node.expr)
        return ref.deref(node.sloc)

    def _eval_addrof(self, node):
        ref = self._eval(node.expr)
        return ref.addrof(node.sloc)

    def _eval_length(self, node):
        val = self._eval(node.array)
        return val.length(self, node.type, node.sloc)

    def _eval_access(self, node):
        idx = self._eval(node.index).native_value()
        val = self._eval(node.array)
        return val.index(idx, node.sloc)

    def _eval_new(self, node):
        if node.is_array:
            return construct(self, node.sloc, node.type, HEAP)
        else:
            assert(is_pointer(node.type))
            e = construct(self, node.sloc, node.type.inner_type, HEAP)
            s = Storage(node.sloc, e, HEAP)
            return Pointer(node.sloc, s)

    def _eval_delete(self, node):
        ref = self._eval(node.expr)
        if node.is_array:
            ref.delete_array(node.sloc)
        else:
            ref.delete(node.sloc)

    def _eval_assert(self, node):
        if not self._eval(node.cond).native_value():
            raise_error("assertion failed", node.sloc)

    def _eval_assume(self, node):
        pass # NOTE: handle this case if necessary

    def _eval_show(self, node):
        print(node.message)

    def _eval_print(self, node):
        val = self._eval(node.value)
        print(repr(val))

    # kernel interactions

    def _block(self, sloc):
        raise ThreadBlocked(self.running, sloc)

    def _eval_start_sim(self, node):
        maxtime = float('inf')
        if node.steps is not None:
            maxtime = self._eval(node.steps).native_value()
        self.kernel.simulation_cycle(maxtime)

    def _eval_stop_sim(self, node):
        raise_error('not implemented', node.sloc)

    def _eval_notify_event(self, node):
        if node.delay is not None:
            n = self._eval(node.delay).native_value()
            self.kernel.timed_notify(node.event, n)
        else:
            self.kernel.immediate_notify(node.event)

    def _eval_wait_event(self, node):
        self.kernel.wait_event(node.event)
        self._block(node.sloc)

    def _eval_wait_time(self, node):
        self.kernel.wait_time(self._eval(node.delay).native_value())
        self._block(node.sloc)

    def _eval_request_update(self, node):
        self.kernel.request_update(node.update)

    def _eval_dont_initialize(self, node):
        raise_error("not implemented", node.sloc)

    def _eval_suspend(self, node):
        self.kernel.suspend(node.thread)

    def _eval_resume(self, node):
        self.kernel.resume(node.thread)


def simulate(prog):
    vm = Interpreter(prog)
    vm.run()
    kernel = vm.kernel

    def str_seq(xs):
        return [e.name for e in xs]

    def str_dict(d):
        return {k.name:str_seq(v) for k, v in d.iteritems()}

    print(":: finished simulation")
    print("  time: {}".format(kernel.sim_time))
    print("  runnable :  {}".format(str_seq(kernel.runnable)))
    print("  suspended:  {}".format(str_seq(kernel.suspended)))
    print("  terminated: {}".format(str_seq(kernel.terminated)))
    print("  pending events:  {}".format(str_dict(kernel.pending)))
    print("  waiting threads: {}".format(str_dict(kernel.waiting)))
    print("  delayed threads: {}".format(str_dict(kernel.delayed)))